/********************************************************************\
 * action.c --  Handler for a button click.                         *
 * Copyright (C) 2002 Linas Vepstas <linas@linas.org>               *
 * http://dwi.sourceforge.net                                       *
 *                                                                  *
 * This library is free software; you can redistribute it and/or    *
 * modify it under the terms of the GNU Lesser General Public       *
 * License as published by the Free Software Foundation; either     *
 * version 2.1 of the License, or (at your option) any later version.
 *                                                                  *
 * This library is distributed in the hope that it will be useful,  *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of   *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the    *
 * GNU Lesser General Public License for more details.              *
 *                                                                  *
 * You should have received a copy of the GNU Lesser General Public *
 * License along with this program; if not, contact:                *
 *                                                                  *
 * Free Software Foundation           Voice:  +1-617-542-5942       *
 * 59 Temple Place - Suite 330        Fax:    +1-617-542-2652       *
 * Boston, MA  02111-1307,  USA       gnu@gnu.org                   *
\********************************************************************/

/*
 * FUNCTION:
 * Handler for a button click.
 *
 * HISTORY:
 * Linas Vepstas March 2002
 */

#include "config.h"

#include <string.h>

#include <glib.h>
#include <gnome.h>
#include <gtk/gtk.h>

#include "action.h"
#include "duifield.h"
#include "duifield-gobj.h"
#include "duifield-gtk.h"
#include "duifield-hash.h"
#include "duifieldmap.h"
#include "duifilter.h"
#include "duiresolver.h"
#include "duitxnquery.h"
#include "perr.h"
#include "report.h"
#include "signal.h"
#include "window.h"

typedef struct DuiCheck_s DuiCheck;
typedef struct DuiRefresh_s DuiRefresh;

struct DuiAction_s
{
	DuiWindow *window;  /* the window in which we live */
	char * name;        /* our name */
	gboolean is_finalized;

	/* SQL database object */
   DuiTxnQuery *query;

	/* List of non-table targets (the target field is not an sql table) */
	GList *non_table_targets;

	/* The report that will report the results of the query */
	char * report_name;
	DuiReport * report;
	char * row_name;

	/* List of gtk signals that can trigger this action. */
	GList *signals;

	/* List of checks that must be performed first (and pass)
	 * before the form is allowed to proceed.
	 */
	GList *check_list;

	/* List of windows that need to be refreshed by this action */
	GList *refresh_list;

	/* name resolution infrastructure */
	DuiResolver * resolver;
};

struct DuiRefresh_s
{
	char * reportname;
	DuiReport *report;
};

struct DuiCheck_s
{
	char * actionname;
	DuiAction *action;
};

/* =================================================================== */

DuiAction * 
dui_action_new (DuiWindow *win, const char * name)
{
	DuiAction *act;

	if (!win) return NULL;

	act = g_new (DuiAction, 1);
	act->window = win;
	act->name = g_strdup (name);
	act->is_finalized = FALSE;

	act->signals = NULL;

	act->resolver = dui_resolver_new();
   act->query = dui_txnquery_new();
	dui_txnquery_set_resolver (act->query, act->resolver);

	act->non_table_targets = NULL;

	act->report_name = NULL;
	act->report = NULL;
	act->row_name = NULL;
	act->check_list = NULL;
	act->refresh_list = NULL;

	dui_window_add_action (win, act);
	return act;
}

/* =================================================================== */

void 
dui_action_destroy (DuiAction * act)
{
	GList *node;
	if (!act) return;

	act->window = NULL;
	g_free (act->name);
	act->name = NULL;

	dui_txnquery_destroy (act->query);

	/* remove  terms */
	for (node=act->non_table_targets; node; node=node->next)
	{
		DuiFieldMap *fm = node->data;
		dui_field_map_destroy (fm);
	}
	g_list_free (act->non_table_targets);
	act->non_table_targets = NULL;

	g_free(act->report_name);
	g_free(act->row_name);
	act->report_name = NULL;
	act->report = NULL;
	act->row_name = NULL;

	/* the signals themselves will be reaped by window */
	g_list_free (act->signals);
	act->signals = NULL;

	for (node=act->check_list; node; node=node->next)
	{
		DuiCheck *chk = node->data;
		g_free (chk->actionname);
		g_free (chk);
	}
	for (node=act->refresh_list; node; node=node->next)
	{
		DuiRefresh *rsh = node->data;
		g_free (rsh->reportname);
		g_free (rsh);
	}
	g_list_free (act->refresh_list);

	dui_resolver_destroy (act->resolver);
	g_free (act);
}

/* =================================================================== */

const char * 
dui_action_get_name (DuiAction *act)
{
	if (!act) return NULL;
	if (act->name) return (act->name);
	return dui_window_get_name (act->window);
}

/* =================================================================== */

void 
dui_action_set_report (DuiAction * act, const char * dpyname, 
                       const char * row)
{
	if (!act) return;
	if (act->report_name) g_free (act->report_name);
	act->report_name = g_strdup (dpyname);
	act->row_name = g_strdup (row);
}
	
/* =================================================================== */

void 
dui_action_add_chain (DuiAction *act, const char * actname)
{
	DuiCheck *chk;
	if (!act) return;
	
	chk = g_new (DuiCheck, 1);
	chk->actionname = g_strdup (actname);
	chk->action = NULL;

	act->check_list = g_list_prepend (act->check_list, chk);
}

/* =================================================================== */

void 
dui_action_add_refresh (DuiAction *act, const char * name)
{
	DuiRefresh *rsh;
	if (!act) return;
	
	rsh = g_new (DuiRefresh, 1);
	rsh->reportname = g_strdup (name);
	rsh->report = NULL;

	act->refresh_list = g_list_prepend (act->refresh_list, rsh);
}

/* =================================================================== */

void 
dui_action_set_database (DuiAction * act, DuiDatabase *db)
{
	if (!act) return;
   dui_txnquery_set_database (act->query, db);
}
	
DuiDatabase *
dui_action_get_database (DuiAction * act)
{
	if (!act) return NULL;
	return dui_txnquery_get_database (act->query);
}

/* =================================================================== */

void 
dui_action_add_signal (DuiAction *act, const char *objectname, const char *signalname)
{
	DuiSignal *sig;

	if (!act || !objectname || !signalname) return;
	
	sig = dui_signal_new (act->window, objectname, signalname, "submit_form");
	act->signals = g_list_append (act->signals, sig);
}
			
/* =================================================================== */
/* XXX abolosh the pass-through */

void 
dui_action_set_table (DuiAction *act, DuiField *tables, 
                                      const char * sql_querytype)
{
	if (!act || !sql_querytype) return;
	dui_txnquery_set_table (act->query, tables, sql_querytype);
}

/* =================================================================== */

void 
dui_action_add_term (DuiAction * act, DuiFieldMap *fm)
{
	if (!act || !fm) return;

	if (DUI_FIELD_IS_TYPE(&fm->target, DUI_FIELD_SQL) ||
	    DUI_FIELD_IS_TYPE(&fm->target, DUI_FIELD_WHERE) ||
	    DUI_FIELD_IS_TYPE(&fm->target, DUI_FIELD_CONST))
	{
		dui_txnquery_add_term (act->query, fm);
	}
	else
	{
		dui_resolver_add_field (act->resolver, &fm->source);
		dui_resolver_add_field (act->resolver, &fm->target);
		act->non_table_targets = g_list_append (act->non_table_targets, fm);
	}
}

/* =================================================================== */
/* "final_setup" resolves window names, filter names, etc, without
 * actually touching the GUI in any way.
 */

static void
resolve_widget_again (DuiAction *act)
{
	GList *node;
	for (node=act->non_table_targets; node; node=node->next)
	{
		DuiFieldMap *fm = node->data;
		dui_field_map_resolve (fm);
	}
   dui_resolver_resolve_widgets (act->resolver, act->window);
}

static void
final_setup (DuiAction *act)
{
	GList *node;
	DuiInterface *dui;

	if (act->is_finalized) return;
	act->is_finalized = TRUE;
	ENTER ("(act=%p)", act);
	
	/* Locate the reports the action will use, and plug them in. */
	dui = dui_window_get_interface (act->window);
	act->report = dui_interface_find_report_by_name (dui, act->report_name);
	if (act->report_name && !act->report) 
	{
		const char * tmp = dui_action_get_name(act);
		SYNTAX ("In form \'%s\', can't find the report \'%s\'", 
		                tmp, act->report_name);
	}

	/* Do the same thing for the check list */
	for (node=act->check_list; node; node=node->next)
	{
		DuiCheck *chk = node->data;
		chk->action = dui_interface_find_action_by_name (dui, chk->actionname);

		if (chk->actionname && !chk->action) 
		{
			const char * tmp = dui_action_get_name(act);
			SYNTAX ("In form \'%s\', can't find the chain form \'%s\'", 
			                tmp, chk->actionname);
		}
	}
	
	/* Do the same thing for the refresh list */
	for (node=act->refresh_list; node; node=node->next)
	{
		DuiRefresh *rsh = node->data;
		rsh->report = dui_interface_find_report_by_name (dui, rsh->reportname);

		if (rsh->reportname && !rsh->report) 
		{
			const char * tmp = dui_action_get_name(act);
			SYNTAX ("In form \'%s\', can't find the refresh report \'%s\'", 
			                tmp, rsh->reportname);
		}
	}
	
	for (node=act->non_table_targets; node; node=node->next)
	{
		DuiFieldMap *fm = node->data;
		dui_field_map_resolve (fm);
	}

   dui_resolver_resolve_widgets (act->resolver, act->window);
   dui_resolver_resolve_gobj (act->resolver, dui);
   dui_resolver_resolve_hash (act->resolver, dui);
	dui_txnquery_do_realize (act->query);

	LEAVE ("(act=%p)", act);
}

static void
dui_action_signal_handler (GtkObject *obj, DuiAction *act)
{
	dui_action_run (act);
}

void
dui_action_do_realize (DuiAction *act)
{
	GList *node;
	if (!act) return;

	ENTER ("(act=%p)", act);
	/* Don't bother if the parent window doesn't exist. */
	/* Technically, this is an error, it shouldn't happen ... */
	if (0 == dui_window_is_realized(act->window)) return;

	final_setup (act);

	for (node=act->signals; node; node=node->next)
	{
		GtkWidget *w;
		DuiSignal *sig = node->data;
		w = dui_window_get_widget (act->window, dui_signal_get_widget(sig));
		/* might not be a widget if the root window doesn't exist */
		if (!w) continue;  

		dui_signal_connect (sig, GTK_OBJECT(w),
		         GTK_SIGNAL_FUNC(dui_action_signal_handler), act);
	}

	LEAVE ("(act=%p)", act);
}

/* =================================================================== */

int
dui_action_is_realized (DuiAction *act)
{
	if (!act) return 0;

	return dui_window_is_realized(act->window);
}

/* =================================================================== */

void 
dui_action_db_connect (DuiAction *act)
{
	if (!act) return;
	dui_txnquery_connect (act->query);
}

/* =================================================================== */

int
dui_action_run (DuiAction *act)
{
	DuiInterface *dui;
	DuiDBRecordSet *recs = NULL;
	GList *node;
	int rc = 0;

	if (!act) return 0;
	ENTER ("(act=%p \'%s\')", act, dui_action_get_name (act));
	final_setup (act);

	/* We need to resolve the widgets again, because the action window
	 * may have been closed. When it is re-realized, there will be all
	 * new widgets for the thing.  It would probably be more efficient
	 * (and more complex) to catch the window-close signal. Oh well. */
	resolve_widget_again (act);

	/* Loop over chained actions; quit if one of them handled things */
	for (node=act->check_list; node; node=node->next)
	{
		DuiCheck *chk = node->data;
		rc = dui_action_run (chk->action);
		if (rc) return rc;
	}

	/* Loop over list of field maps that don't involve any SQL tables,
	 * and move the data from the source to the target. */
	for (node = act->non_table_targets; node; node=node->next)
	{
		DuiFieldMap *fm = node->data;
		dui_field_map_transfer_data (fm);
	}
	
	dui = dui_window_get_interface (act->window);
	recs = dui_txnquery_run (act->query);

	/* now report the results */
	dui_report_set_last_action (act->report, act, act->row_name);
	rc = dui_report_show_data (act->report, recs, act->row_name);

	/* Don't need it no more */
	dui_recordset_free (recs);

	/* Refresh all windows that show data modified by this query */
	for (node=act->refresh_list; node; node=node->next)
	{
		DuiRefresh *rsh = node->data;
		dui_report_refresh (rsh->report);
	}
	LEAVE ("(act=%p \'%s\') rc=%d", act, dui_action_get_name (act), rc);
	return rc;
}

/* =================================================================== */

void
dui_action_rerun_last_query (DuiAction *act)
{
	DuiDBRecordSet *recs = NULL;

	if (!act) return;

	ENTER ("(act=%p \'%s\')", act, dui_action_get_name (act));

	/* Issue same query as before, and display the results */
	recs = dui_txnquery_rerun_last_query (act->query);
	dui_report_show_data (act->report, recs, act->row_name);
	dui_recordset_free (recs);
	
	LEAVE ("(act=%p \'%s\')", act, dui_action_get_name (act));
}

/* ========================== END OF FILE ============================ */
